#include <iostream>
#include <X11/Xlib.h>
#ifdef WITH_XRANDR
#include <X11/extensions/Xrandr.h>
#endif
#include <msp/io/print.h>
#include <msp/strings/format.h>
#include <msp/strings/lexicalcast.h>
#include "display.h"
#include "display_private.h"

using namespace std;

namespace {

bool error_flag = false;
std::string error_msg;

int x_error_handler(Display *display, XErrorEvent *event)
{
	char err[128];
	XGetErrorText(display, event->error_code, err, sizeof(err));

	string request_code = Msp::lexical_cast<string, int>(event->request_code);
	char req[128];
	XGetErrorDatabaseText(display, "XRequest", request_code.c_str(), request_code.c_str(), req, sizeof(req));

	string msg = Msp::format("Request %s failed with %s [%08X]", req, err, event->resourceid);
	if(error_flag)
		Msp::IO::print(Msp::IO::cerr, "Discarding error: %s\n", msg);
	else
	{
		Msp::IO::print(Msp::IO::cerr, "%s\n", msg);
		error_msg = msg;
		error_flag = true;
	}

	return 0;
}

#ifdef WITH_XRANDR
inline Msp::Graphics::VideoRotation rotation_from_sys(Rotation r)
{
	switch(r)
	{
	case RR_Rotate_90: return Msp::Graphics::ROTATE_RIGHT;
	case RR_Rotate_180: return Msp::Graphics::ROTATE_INVERTED;
	case RR_Rotate_270: return Msp::Graphics::ROTATE_LEFT;
	default: return Msp::Graphics::ROTATE_NORMAL;
	}
}

inline Rotation rotation_to_sys(Msp::Graphics::VideoRotation r)
{
	switch(r)
	{
	case Msp::Graphics::ROTATE_RIGHT: return RR_Rotate_90;
	case Msp::Graphics::ROTATE_INVERTED: return RR_Rotate_180;
	case Msp::Graphics::ROTATE_LEFT: return RR_Rotate_270;
	default: return RR_Rotate_0;
	}
}

inline bool monitor_x_compare(const Msp::Graphics::Monitor &m1, const Msp::Graphics::Monitor &m2)
{
	if(m1.desktop_mode && !m2.desktop_mode)
		return true;
	return m1.x<m2.x;
}
#endif

inline unsigned mode_width(const Msp::Graphics::VideoMode &m, Msp::Graphics::VideoRotation r)
{
	if(r==Msp::Graphics::ROTATE_RIGHT || r==Msp::Graphics::ROTATE_LEFT)
		return m.height;
	else
		return m.width;
}

}


namespace Msp {
namespace Graphics {

Display::Display(const string &disp_name):
	primary_monitor(0),
	priv(new Private)
{
	if(disp_name.empty())
		priv->display = XOpenDisplay(0);
	else
		priv->display = XOpenDisplay(disp_name.c_str());
	if(!priv->display)
		throw runtime_error("XOpenDisplay");

	XSetErrorHandler(x_error_handler);

	err_dialog = new ErrorDialog(this);

#ifdef WITH_XRANDR
	int event_base;
	int error_base;
	if(XRRQueryExtension(priv->display, &event_base, &error_base))
	{
		int major, minor;
		XRRQueryVersion(priv->display, &major, &minor);
		if(major>1 || (major==1 && minor>=2))
		{
			WindowHandle root = DefaultRootWindow(priv->display);
			XRRScreenResources *res = XRRGetScreenResources(priv->display, root);
			RROutput primary = XRRGetOutputPrimary(priv->display, root);

			map<RRMode, XRRModeInfo *> modes_by_id;
			for(int i=0; i<res->nmode; ++i)
				modes_by_id[res->modes[i].id] = &res->modes[i];

			for(int i=0; i<res->noutput; ++i)
			{
				XRROutputInfo *output = XRRGetOutputInfo(priv->display, res, res->outputs[i]);
				XRRCrtcInfo *crtc = (output->crtc ? XRRGetCrtcInfo(priv->display, res, output->crtc) : 0);

				monitors.push_back(Monitor());
				Monitor &monitor = monitors.back();
				monitor.index = monitors.size()-1;
				monitor.name.assign(output->name, output->nameLen);
				priv->monitors.push_back(res->outputs[i]);

				if(crtc)
				{
					monitor.desktop_rotation = rotation_from_sys(crtc->rotation);
					monitor.current_rotation = monitor.desktop_rotation;
					monitor.x = crtc->x;
					monitor.y = crtc->y;
				}

				if(res->outputs[i]==primary)
					primary_monitor = &monitor;

				for(int j=0; j<output->nmode; ++j)
				{
					map<RRMode, XRRModeInfo *>::iterator k = modes_by_id.find(output->modes[j]);
					if(k==modes_by_id.end())
						continue;

					XRRModeInfo *info = k->second;

					VideoMode mode(info->width, info->height);
					mode.index = modes.size();
					mode.monitor = &monitor;
					mode.rate = static_cast<float>(info->dotClock)/(info->hTotal*info->vTotal);
					if(find_mode(mode, 0.01f))
						continue;

					modes.push_back(mode);
					priv->modes.push_back(info->id);
					monitor.video_modes.push_back(&modes.back());

					if(crtc && info->id==crtc->mode)
					{
						monitor.desktop_mode = &modes.back();
						monitor.current_mode = monitor.desktop_mode;
					}
				}

				XRRFreeOutputInfo(output);
				if(crtc)
					XRRFreeCrtcInfo(crtc);
			}

			XRRFreeScreenResources(res);

			monitors.sort(monitor_x_compare);
			Monitor *prev_enabled = 0;
			for(list<Monitor>::iterator i=monitors.begin(); i!=monitors.end(); ++i)
				if(i->desktop_mode)
				{
					i->next_left = prev_enabled;
					if(prev_enabled)
						prev_enabled->next_right = &*i;
					prev_enabled = &*i;
				}

			if(!primary_monitor || !primary_monitor->desktop_mode)
			{
				// XRandR didn't give a sensible primary monitor.  Try to guess one.
				unsigned largest = 0;
				for(list<Monitor>::iterator i=monitors.begin(); i!=monitors.end(); ++i)
					if(i->desktop_mode)
					{
						unsigned size = i->desktop_mode->width*i->desktop_mode->height;
						if(size>largest)
						{
							largest = size;
							primary_monitor = &*i;
						}
					}
			}
		}
	}
#endif
}

Display::~Display()
{
	XCloseDisplay(priv->display);
	delete priv;
	delete err_dialog;
}

void Display::set_mode(const VideoMode &requested_mode, bool exclusive)
{
#ifdef WITH_XRANDR
	const VideoMode *mode = find_mode(requested_mode);
	if(!mode)
		throw unsupported_video_mode(requested_mode);

	VideoRotation requested_rotation = requested_mode.rotation;
	if(requested_rotation==ROTATE_ANY)
		requested_rotation = mode->monitor->desktop_rotation;

	WindowHandle root = DefaultRootWindow(priv->display);
	XRRScreenResources *res = XRRGetScreenResources(priv->display, root);
	RROutput output = priv->monitors[mode->monitor->index];
	XRROutputInfo *output_info = XRRGetOutputInfo(priv->display, res, output);

	// Check if the output already has a CRTC and find a free one if it doesn't
	RRCrtc crtc = output_info->crtc;
	XRRCrtcInfo *crtc_info = 0;
	if(crtc)
		crtc_info = XRRGetCrtcInfo(priv->display, res, crtc);
	else
	{
		for(int i=0; i<res->ncrtc; ++i)
		{
			crtc_info = XRRGetCrtcInfo(priv->display, res, res->crtcs[i]);
			if(!crtc_info->noutput)
			{
				crtc = res->crtcs[i];
				break;
			}
			XRRFreeCrtcInfo(crtc_info);
		}

		if(!crtc)
		{
			XRRFreeOutputInfo(output_info);
			throw unsupported_video_mode(requested_mode);
		}
	}

	int x = 0;
	int y = 0;

	if(exclusive)
	{
		// Disable other outputs for exclusive mode
		for(list<Monitor>::iterator i=monitors.begin(); i!=monitors.end(); ++i)
			if(&*i!=mode->monitor)
			{
				XRROutputInfo *o = XRRGetOutputInfo(priv->display, res, priv->monitors[i->index]);
				if(o->crtc)
					XRRSetCrtcConfig(priv->display, res, o->crtc, CurrentTime, 0, 0, 0, RR_Rotate_0, 0, 0);
				XRRFreeOutputInfo(o);

				i->current_mode = 0;
				i->current_rotation = ROTATE_NORMAL;
				i->x = 0;
				i->y = 0;
			}
	}
	else
	{
		const Monitor *left = mode->monitor->next_left;
		while(left && !left->current_mode)
			left = left->next_left;

		if(left)
		{
			x = left->x+mode_width(*left->current_mode, left->current_rotation);
			y = left->y;
		}
	}

	XRRSetCrtcConfig(priv->display, res, crtc, CurrentTime, x, y, priv->modes[mode->index], rotation_to_sys(requested_rotation), &output, 1);

	list<Monitor>::iterator i;
	for(i=monitors.begin(); i!=monitors.end(); ++i)
		if(&*i==mode->monitor)
		{
			i->current_mode = mode;
			i->current_rotation = requested_rotation;
			i->x = x;
			i->y = y;

			x += mode_width(*mode, requested_rotation);
			++i;
			break;
		}

	for(; i!=monitors.end(); ++i)
		if(i->current_mode)
		{
			XRROutputInfo *o = XRRGetOutputInfo(priv->display, res, priv->monitors[i->index]);
			XRRSetCrtcConfig(priv->display, res, o->crtc, CurrentTime, x, y, priv->modes[i->current_mode->index], rotation_to_sys(i->current_rotation), &priv->monitors[i->index], 1);

			XRRPanning panning;
			panning.timestamp = CurrentTime;
			panning.left = x;
			panning.top = y;
			panning.width = i->current_mode->width;
			panning.height = i->current_mode->height;
			panning.track_left = panning.left;
			panning.track_top = panning.top;
			panning.track_width = panning.width;
			panning.track_height = panning.height;
			panning.border_left = 0;
			panning.border_top = 0;
			panning.border_right = 0;
			panning.border_bottom = 0;
			XRRSetPanning(priv->display, res, o->crtc, &panning);

			XRRFreeOutputInfo(o);

			i->x = x;
			i->y = y;

			x += mode_width(*i->current_mode, i->current_rotation);
		}

	XRRFreeOutputInfo(output_info);
	XRRFreeCrtcInfo(crtc_info);
	XRRFreeScreenResources(res);
#else
	(void)requested_mode;
	(void)exclusive;
	throw runtime_error("no xrandr support");
#endif
}

bool Display::process_events()
{
	int pending = XPending(priv->display);
	if(pending==0)
		return false;

	for(; pending--;)
	{
		Window::Event event;
		XNextEvent(priv->display, &event.xevent);

		check_error();

		map<WindowHandle, Window *>::iterator j = priv->windows.find(event.xevent.xany.window);
		if(j!=priv->windows.end())
		{
			/* Filter keyboard autorepeat.  If this packet is a KeyRelease and
			the next one is a KeyPress with the exact same parameters, they
			indicate autorepeat and must be dropped. */
			if(event.xevent.type==KeyRelease && !j->second->get_keyboard_autorepeat() && pending>0)
			{
				XKeyEvent &kev = event.xevent.xkey;
				XEvent ev2;
				XPeekEvent(priv->display, &ev2);
				if(ev2.type==KeyPress)
				{
					XKeyEvent &kev2 = ev2.xkey;
					if(kev2.window==kev.window && kev2.time==kev.time && kev2.keycode==kev.keycode)
					{
						XNextEvent(priv->display, &ev2);
						--pending;
						continue;
					}
				}
			}

			j->second->event(event);
		}
	}

	return true;
}

void Display::check_error()
{
	if(error_flag)
	{
		error_flag = false;
		throw runtime_error(error_msg);
	}
}

} // namespace Msp
} // namespace Graphics
